
import java.util.*;
import java.text.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.*;
import javax.swing.event.*;


class Monster {
    int x,y;
}

class Critter {
    int x,y;
    int numRedGenes = 0; 
    // The value should be between 0 and maxGenes. This is what
    // determines survival - the more the better.
}



public class Evolution extends JPanel {
        
    int xWidth = 500;                // Currently, we're limiting the landscape.
    int yWidth = 500;
    int numMonsters = 10;            // Fixed number of monsters, but critters will grow.
    int numCritters = 10;
    int maxCritters = 20;            // Max# critters allowed.
    int monsterSize = 30;            // The monster is bigger, naturally.
    int critterSize = 15;
    int time = 0;                    // Current time.
    int t = 1000;                    // Time limit.

    int attackDistance = 45;         // A critter has to be this close to be attacked.
    int randomMoveDist = 20;         // Amount to move in random direction.

    int geneThreshold = 8;           // A critter with this much color survives attack.
    int maxGenes = 10;               // The maximum number of color levels.
    int mutationSpeed = 1;           // Larger jumps in color are allowed.
    int birthRate = 2;               // Each surviving critter gives birth to this many.
    double deathProb = 0;

    LinkedList<Monster> monsters;    // Data structures to hold the monsters and critters.
    LinkedList<Critter> critters;
    
    // GUI variables.
    JTextField speedField = new JTextField (5);
    JTextField mField = new JTextField (3);
    JTextField cField = new JTextField (3);
    JTextField bField = new JTextField (3);
    JTextField dField = new JTextField (4);
    int sleepTime = 100;
    String currentMsg = "";
    Thread currentThread;



    // When the "go" button is pressed, make a thread and fire off a simulation run.

    void go ()
    {
        if (currentThread != null) {
            currentThread.interrupt ();
            currentThread = null;
        }
        
	currentThread = new Thread () {
		public void run () 
		{
		    simulate ();
		}
	    };
	currentThread.start ();
    }
    


    // Perform one simulation run.

    void simulate ()
    {
        reset ();
        while (time <= t) {  
	    try {
		Thread.sleep (sleepTime);
	    }
	    catch (InterruptedException e) {
                break;
	    }

            nextStep ();
        }
	repaint ();
    }
    
    // Initialize variables at t=0

    void reset ()
    {
        // Read m, c and b.
        try {
            mutationSpeed = Integer.parseInt (speedField.getText());
            numMonsters = Integer.parseInt (mField.getText());
            numCritters = Integer.parseInt (cField.getText());
            birthRate = Integer.parseInt (bField.getText());
            deathProb = Double.parseDouble (dField.getText());
        }
        catch (NumberFormatException e) {
            mutationSpeed = 1;
            numMonsters = 10;
            numCritters = 10;
            birthRate = 3;
        }

	// Reset.
        monsters = new LinkedList<Monster> ();
        for (int i=0; i<numMonsters; i++) {
            int x = UniformRandom.uniform (0, 500);
            int y = UniformRandom.uniform (0, 500);
            Monster m = new Monster ();
            m.x = x;
            m.y = y;
            monsters.add (m);
        }
        
        critters = new LinkedList<Critter> ();
        for (int j=0; j<numCritters; j++) {
            int x = UniformRandom.uniform (0, 500);
            int y = UniformRandom.uniform (0, 500);
            Critter c = new Critter ();
            c.x = x;
            c.y = y;
            critters.add (c);
        }

        time = 0;
        repaint ();
    }
    

    // Computation at each step.

    synchronized void nextStep ()
    {
        // Update current time.
        time += 1;
        currentMsg = "time: " + time;

        // See if any critters get eaten.
        LinkedList<Critter> deadCritters = new LinkedList<Critter> ();
        for (Critter c: critters) {
            if (c.numRedGenes >= geneThreshold) {
                // Unaffected by attack.
                continue;
            }
            
            boolean safe = true;
            for (Monster m: monsters) {
                if ( distance (c.x,c.y, m.x,m.y) < attackDistance ) {
                    // There is a monster within attack distance.
                    safe = false;
                }
            }
            if (! safe) {
                deadCritters.add (c);
            }
        }

        // We perform removal later because of synchronization conflicts.
        for (Critter c: deadCritters) {
            critters.remove (c);
        }
        

        // Move randomly.
        for (Monster m: monsters) {
            double angle = UniformRandom.uniform (0.0, 2*Math.PI);
            int x = (int) Math.round (randomMoveDist * Math.cos(angle));
            int y = (int) Math.round (randomMoveDist * Math.sin(angle));
            if (withinBounds (m.x+x, m.y+y)) {
                m.x += x;
                m.y += y;
            }
        }
        for (Critter c: critters) {
            double angle = UniformRandom.uniform (0.0, 2*Math.PI);
            int x = (int) Math.round (randomMoveDist * Math.cos(angle));
            int y = (int) Math.round (randomMoveDist * Math.sin(angle));
            if (withinBounds (c.x+x, c.y+y)) {
                c.x += x;
                c.y += y;
            }
        }
        
        // See if population bound is exceeded.
        if (critters.size () > maxCritters) {
            // Remove some.
            deadCritters = new LinkedList<Critter> ();
            for (Critter c: critters) {
                if (UniformRandom.uniform() < deathProb) {
                    deadCritters.add (c);
                }
            }
            // Now remove.
            for (Critter c: deadCritters) {
                critters.remove (c);
            }
            repaint ();
            return;
        }

        // Critters give birth and evolve.
        LinkedList<Critter> babyCritters = new LinkedList<Critter> ();
        for (Critter c: critters) {
            // For each critter, create more.
            for (int j=0; j<birthRate; j++) {
                int amount = UniformRandom.uniform (0, mutationSpeed);
                if (UniformRandom.uniform() < 0.5) {
                    amount = - amount;
                }
                int newValue = c.numRedGenes + amount;
                if (newValue < 0) newValue = 0;
                if (newValue > maxGenes) newValue = maxGenes;
                Critter baby = new Critter ();
                baby.x = 250;
                baby.y = 250;
                baby.numRedGenes = newValue;
                babyCritters.add (baby);
            }
        }

        for (Critter c: babyCritters) {
            critters.add (c);
        }

        repaint ();
    }


    double distance (double x1, double y1, double x2, double y2)
    {
	double d = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
	return Math.sqrt (d);
    }


    boolean withinBounds (int x, int y) 
    {
        if ((x < 0) || (x > 500)) {
            return false;
        }
        if ((y < 0) || (y > 500)) {
            return false;
        }
        return true;
    }
    

    String[] makeStateString ()
    {
        int[] colorHist = new int [maxGenes+1];
        for (Critter c: critters) {
            colorHist [c.numRedGenes] ++;
        }
        String[] str = new String [maxGenes + 1];
        for (int i=0; i<=maxGenes; i++) {
            str[i] = "# with " + i + " genes: " + String.format("%2d",colorHist[i]);
        }
        return str;
    }


    ////////////////////////////////////////////////////////////////////////
    // GUI stuff


    public void paintComponent (Graphics g)
    {
	super.paintComponent (g);
	Dimension D = this.getSize ();

	// Background.
	g.setColor (Color.white);
	g.fillRect (0,0, D.width, D.height);
        
        if ( (critters == null) || (monsters == null) ) {
            return;
        }
        
        // Critters.
        for (Critter c: critters) {
            g.setColor (new Color (20*c.numRedGenes,0,0));
            g.fillOval (c.x-critterSize,D.height-c.y-critterSize, 2*critterSize,2*critterSize);
        }
        
        // Monsters.
	g.setColor (Color.green);
        for (Monster m: monsters) {
            g.fillOval (m.x-monsterSize,D.height-m.y-monsterSize, 2*monsterSize,2*monsterSize);
        }

        // Message.
	g.setColor (Color.black);
	g.drawString (currentMsg, 600, 20);
        String[] stateStr = makeStateString ();
        for (int i=0; i<stateStr.length; i++) {
            g.drawString (stateStr[i], 600, i*20 + 40);
        }
    }
    
    JPanel makeBottomPanel ()
    {
        JPanel panel = new JPanel ();
        
        panel.add (new JLabel ("m: "));
        mField.setText (""+numMonsters);
        panel.add (mField);
        panel.add (new JLabel ("   "));
        panel.add (new JLabel ("c: "));
        cField.setText (""+numCritters);
        panel.add (cField);
        panel.add (new JLabel ("   "));
        panel.add (new JLabel ("b: "));
        bField.setText (""+birthRate);
        panel.add (bField);
        panel.add (new JLabel ("   "));
        panel.add (new JLabel ("d: "));
        dField.setText (""+deathProb);
        panel.add (dField);
        panel.add (new JLabel ("     "));
        panel.add (new JLabel ("Mutation speed:"));
        speedField.setText (""+mutationSpeed);
        panel.add (speedField);

        panel.add (new JLabel ("     "));

        panel.add (new JLabel ("  "));
	JButton resetB = new JButton ("Reset");
	resetB.addActionListener (
	   new ActionListener () {
		   public void actionPerformed (ActionEvent a)
		   {
		       reset ();
		   }
           }
        );
	panel.add (resetB);

        panel.add (new JLabel ("  "));
	JButton nextB = new JButton ("Next");
	nextB.addActionListener (
	   new ActionListener () {
		   public void actionPerformed (ActionEvent a)
		   {
		       nextStep ();
		   }
           }
        );
	panel.add (nextB);

        panel.add (new JLabel ("  "));
	JButton goB = new JButton ("Go");
	goB.addActionListener (
	   new ActionListener () {
		   public void actionPerformed (ActionEvent a)
		   {
		       go ();
		   }
           }
        );
	panel.add (goB);

        panel.add (new JLabel ("  "));
	JButton quitB = new JButton ("Quit");
	quitB.addActionListener (
	   new ActionListener () {
		   public void actionPerformed (ActionEvent a)
		   {
		       System.exit(0);
		   }
           }
        );
	panel.add (quitB);
        
        return panel;
    }


    void makeFrame ()
    {
        JFrame frame = new JFrame ();
        frame.setSize (950, 700);
        frame.setTitle ("Evolution");
        Container cPane = frame.getContentPane();
        cPane.add (makeBottomPanel(), BorderLayout.SOUTH);
        cPane.add (this, BorderLayout.CENTER);
        frame.setVisible (true);
    }


    public static void main (String[] argv)
    {
        Evolution e = new Evolution ();
        e.makeFrame ();
    }

}
